# -*- coding: utf-8 -*-
#
# File: replicator.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# GNU General Public License (GPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from zope.interface import implements
from zope import event

import csv, cStringIO

from DateTime import DateTime
from time import strptime

from Products.CMFCore.utils import getToolByName

from Products.Archetypes.event import ObjectInitializedEvent
from Products.Archetypes.event import ObjectEditedEvent

from interfaces import Icsvreplicata
#from config import getHandlers, default_handler

from exceptions import *
from Products.csvreplicata import getPortalTypes

import logging
logger = logging.getLogger('CSV REPLICATOR')

class Replicator(object):
    """ a folder able to import/export its content as CSV 
    """
    implements(Icsvreplicata)

    def __init__(self, context):
        """Initialize adapter."""
        self.context = context
        self.flag = True

    def csvimport(self, csvfile, encoding=None, delimiter=None,
                  stringdelimiter=None, datetimeformat=None,
                  conflict_winner="SERVER", wf_transition=None, zip=None,
                  vocabularyvalue="No", count_created=0, count_modified=0,
                  errors=[]):
        """
        CSV import.
        
        Calls recursively self._csvimport while self.flag
        """
        while self.flag :
            
            count_created, count_modified, export_date, errors = \
                    self._csvimport(csvfile, encoding, delimiter,
                                    stringdelimiter, datetimeformat,
                                    conflict_winner, wf_transition, zip,
                                    vocabularyvalue, count_created,
                                    count_modified, errors)

        return (count_created, count_modified, export_date, errors)

    def _csvimport(self, csvfile, encoding, delimiter, stringdelimiter,
                   datetimeformat, conflict_winner, wf_transition, zip,
                   vocabularyvalue, count_created, count_modified, errors):
        
            csvfile.seek(0)
        
            # get portal types
            types = getPortalTypes(self.context)
            
            # read parameters
            csvtool = getToolByName(self.context, "portal_csvreplicatatool")
            if encoding is None:
                encoding = csvtool.getEncoding()
            if delimiter is None:
                delimiter = csvtool.getDelimiter()
            if stringdelimiter is None:
                stringdelimiter = csvtool.getStringdelimiter()
            if datetimeformat is None:
                #datetimeformat = csvtool.getDateTimeFormat()
                datetimeformat = "%Y-%m-%d"
            self.datetimeformat = datetimeformat
            self.vocabularyvalue = vocabularyvalue
                
                
            # read csv
            reader = csv.reader(csvfile, delimiter=delimiter,
                                quotechar=stringdelimiter,
                                quoting=csv.QUOTE_NONNUMERIC)
            line = 1
            # parse header
            head = reader.next()
            export_folder = head[0]
            export_date_str = str(head[1])
            export_date = DateTime(int(export_date_str[0:4]),
                                   int(export_date_str[4:6]),
                                   int(export_date_str[6:8]),
                                   int(export_date_str[8:10]),
                                   int(export_date_str[10:12]),
                                   int(export_date_str[12:14]))
            
            # parse content
            specific_fields = None
            label_line = False
            needs_another_loop = False
            for row in reader:
                line = line + 1
                if not label_line:
                    # read type fields
                    if row[0] == "parent":
                        specific_fields = row[3:]
                        # next line must not be considered (labels line)
                        label_line = True
                    # read values
                    else:
                        try:
                            (is_new, is_modified) = \
                                            self.importObject(row,
                                                              specific_fields,
                                                              types[row[2]],
                                                              conflict_winner,
                                                              export_date,
                                                              wf_transition,
                                                              zip)
                            if is_new:
                                count_created = count_created + 1
                            elif is_modified:
                                count_modified = count_modified + 1
    
                        except csvreplicataConflictException, e:
                            #errors.append("Conflict on line " + \
                                          #str(line)+": %s" % (e))
                            
                            # here content is modified during a second
                            # (or more) parsing "
                            pass
                            
                        except csvreplicataBrokenReferenceException, e:
                            needs_another_loop = True

                        except csvreplicataNonExistentContainer, e:
                            needs_another_loop = True
                            pass
    
                        except Exception, e:
                            errors.append("Error in line "+str(line) + \
                                          ": %s" % (e))
                            
                else:
                    label_line = False
                    
            self.flag = needs_another_loop
            
            return (count_created, count_modified, export_date, errors)
        


    def importObject(self, row, specific_fields, type, conflict_winner,
                     export_date, wf_transition, zip):
        """
        """
        modified = False
        is_new_object = False
        protected = True
        parent_path = row[0]
        id = row[1]
        type_class = row[2]
        if parent_path == "":
            container = self.context
        else:
            try:
                container = self.context.unrestrictedTraverse(parent_path)
            except:
                raise csvreplicataNonExistentContainer, \
                "Non existent container %s " % parent_path
            
            
        obj = getattr(container, id, None)
        if obj is None:
            # object does not exist, let's create it
            attool = getToolByName(self.context, 'archetype_tool')
            at = attool.lookupType(type[0], type[1])
            portal_type = at['portal_type']
            container.invokeFactory(portal_type, id=id)
            obj = getattr(container, id)
            is_new_object = True
            protected = False
        else:
            # object exists, so check conflicts
            lastmodified = obj.modified()
            if lastmodified > export_date:
                if conflict_winner == "LOCAL":
                    protected = False
            else:
                protected = False
                
        # update object
        csvtool = getToolByName(self.context, "portal_csvreplicatatool")
        handlers = csvtool.getHandlers()
        i = 3
        for f in specific_fields:
            if f is not None and f != "":
                type = obj.Schema().getField(f).getType()
                h = handlers.get(type, handlers['default_handler'])
                handler = h['handler_class']
                old_value = handler.get(obj, f, context=self)
                if old_value != row[i]:
                    if protected:
                        raise csvreplicataConflictException, \
                        "Overlapping content modified on" \
                        " the server after exportation"
                    else:
                        modified = True
                        if h['file']:
                            handler.set(obj, f, row[i], context=self, zip=zip)
                        else:
                            handler.set(obj, f, row[i], context=self)
            i = i+1
        
        # call events
        if is_new_object:
            event.notify(ObjectInitializedEvent(obj))
            obj.at_post_create_script()
            obj.indexObject()
            try:
                wftool = getToolByName(self.context, 'portal_workflow')
                wftool.doActionFor( obj, wf_transition)
            except Exception:
               pass     
        elif modified:
            event.notify(ObjectEditedEvent(obj))
            obj.at_post_edit_script()
            obj.reindexObject()
    
        return (is_new_object, modified)
    
    def csvexport(self, encoding=None, delimiter=None, stringdelimiter=None,
                  datetimeformat=None, depth=1, wf_states=None, zip=None,
                  vocabularyvalue="No", exportable_content_types=None):
        """
        """
        # read parameters
        csvtool = getToolByName(self.context, "portal_csvreplicatatool")
        if encoding is None:
            encoding = csvtool.getEncoding()
        if delimiter is None:
            delimiter = csvtool.getDelimiter()
        if stringdelimiter is None:
            stringdelimiter = csvtool.getStringdelimiter()
        if datetimeformat is None:
            #datetimeformat = csvtool.getDateTimeFormat()
            datetimeformat = "%Y-%m-%d"
        self.datetimeformat = datetimeformat
        self.vocabularyvalue = vocabularyvalue

        
        # initialize csv
        stream = cStringIO.StringIO()
        writer = csv.writer(stream, delimiter=delimiter,
                            quotechar=stringdelimiter,
                            quoting=csv.QUOTE_NONNUMERIC)
        
        writer.writerow(["/".join\
                         (self.context.getPhysicalPath()).encode(encoding),
                         DateTime().strftime(format='%Y%m%d%H%M%S').\
                         encode(encoding)])
        
        # search objects
        #exportable_content_types = csvtool.getReplicableTypesSorted()
        if exportable_content_types is not None:
            if self.context.Type() == "Smart Folder":
                all = self.context.queryCatalog(full_objects=True)
                exportable_objects = []
                for o in all:
                    if o.Type() in exportable_content_types:
                        exportable_objects.append(o)
            else:
                query = {'portal_type': exportable_content_types}
                if depth == 0:
                    path = {'query':"/".join(self.context.getPhysicalPath())}
                else:
                    path = {'query':"/".join(self.context.getPhysicalPath()),
                            'depth':depth}
                query['path'] = path
                
                if wf_states is not None:
                    query['review_state'] = wf_states
                    
                search_exportable = self.context.portal_catalog.searchResults(
                    query)
                exportable_objects = [o.getObject() for o in search_exportable]
        else:
            exportable_objects = []
        
        # export content
        currenttype = None
        currentfields = []
        for obj in exportable_objects:
            type = str(obj.getTypeInfo().id)
            if not(type == currenttype):
                currenttype = type
                # get type fields
                currentfields = self.getTypeFields(type)
                # write ids
                writer.writerow([s[0].encode(encoding) for s in currentfields])
                # writes labels
                writer.writerow([s[1].encode(encoding) for s in currentfields])
                # store type specific fields list
                current_specific_fields = [f[0] for f in currentfields[3:]]
            
            values = self.getObjectValues(obj, current_specific_fields, zip)
            # write values
            writer.writerow([s.encode(encoding) for s in values])
        
        return stream.getvalue()

    def getTypeFields(self, type):
        """
        """
        types = getPortalTypes(self.context)
        attool = getToolByName(self.context, 'archetype_tool')
        csvtool = getToolByName(self.context, "portal_csvreplicatatool")
        if csvtool.replicabletypes.has_key(type):
            schematas = csvtool.replicabletypes[type]
        else:
            schematas = []
        notExportableFieldClasses = csvtool.getExcludedfieldsclasses()
        notExportableFields = csvtool.getExcludedfields()
        at = attool.lookupType(types[type][0], types[type][1])
        at_class = at['klass']
        types = [('parent', 'Parent folder'), ('id', 'Identifier'),
                ('type','Content type')] 
        for schemata in schematas:
            fields = at_class.schema.getSchemataFields(schemata)
            #TODO: use getTranslationService to get the i18n translation
            types.extend([(f.getName(), f.widget.label) 
                          for f in fields 
                          if f.__class__.__name__ not in \
                          notExportableFieldClasses 
                          and f.getName() not in notExportableFields])
        return types
    
    def getObjectValues(self, obj, specific_fields, zip):
        """
        """
        # compute parent path
        current = "/".join(self.context.getPhysicalPath())
        parent_path = "/".join(obj.getParentNode().getPhysicalPath())
        if parent_path == current:
            parent_path = ""
        elif parent_path.startswith(current+"/"):
            parent_path = parent_path[len(current)+1:]
            
        # add the 3 standard first columns
        values = [parent_path, obj.id, obj.getTypeInfo().id]
        
        csvtool = getToolByName(self.context, "portal_csvreplicatatool")
        handlers = csvtool.getHandlers()
        # add specific columns
        for f in specific_fields:
            type = obj.Schema().getField(f).getType()
            h = handlers.get(type, handlers['default_handler'])
            handler = h['handler_class']
            if h['file']:
                v = handler.get(obj, f, context=self, zip=zip)
            else:
                v = handler.get(obj, f, context=self)
            values.append(v)
        return values



    


# -*- coding: utf-8 -*-
#
# File: file.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# GNU General Public License (GPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'


from zope import interface

from Products.csvreplicata.exceptions import *
from Products.csvreplicata.interfaces import ICSVFile

from base import CSVdefault

import logging
logger = logging.getLogger('HANDLER')
 
        
class CSVFile(CSVdefault):
    """
    """
    interface.implements(ICSVFile)
    
    def get(self, obj, field, context=None, zip=None):
        """
        """
        f = obj.Schema().getField(field).get(obj)
        if not f :
            return ''
        else:
            filename = f.filename
            if zip is not None:
                #logger.error(obj.Schema().getField(field).getType())
                if obj.Schema().getField(field).getType() in \
                ("plone.app.blob.subtypes.file.ExtensionBlobField",
                 "Products.Archetypes.Field.FileField",
                 "Products.Archetypes.Field.ImageField"):
                    
                    f = str(f.data)
                zip.writestr(filename, f)
            return filename
    
    def set(self, obj, field, value, context=None, zip=None):
        if value == '':
            raise csvreplicataException, "No filename for %s" % (field)
        elif zip is None:
            raise csvreplicataException, "No zip file provided"
        else:
            if obj.Schema().getField(field).writeable(obj):
                obj.Schema().getField(field).set(obj, zip.read(value))
                obj.Schema().getField(field).get(obj).setFilename(value)
            else:
                raise csvreplicataPermissionException, \
                "Insufficient privileges to modify this object and/or field"
        


# -*- coding: utf-8 -*-

from zope.interface import Interface


class ICSVDefault(Interface):
    """CSV Default.
    
    A CSV Field defaut Interface
    """
    
    def get(obj, field, context=None):
        """A getter."""
    
    def set(obj, field, value, context=None):
        """ A setter."""
    
    def store(field, obj, value):
        """Store field value."""


class ICSVFile(Interface):
    """CSV File.
    
    A CSV File defaut Interface
    """
    
    def get(obj, field, context=None, zip=None):
        """A getter."""
    
    def set(obj, field, value, context=None, zip=None):
        """ A setter."""


class IcsvreplicataTool(Interface):
    """Marker interface for .csvreplicataTool.csvreplicataTool
    """


class ICSVReplicable(Interface):
    """Marker interface for replicable folders
    """
    
class Icsvreplicata(Interface):
    """ adapter interface
    """



from setuptools import setup, find_packages
import os

version = '1.0.6'

setup(name='Products.csvreplicata',
      version=version,
      description="CSV import/export for Archetypes",
      long_description='\n'.join(
        open(os.path.join(*path)).read() for path in [
            ("README.txt",),
            ("docs", "HISTORY.txt"),
            ("Products", "csvreplicata", "tests", "csvreplicata.txt")]),
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='archetypes, import/export, csv, replication, synchronisation',
      author='Eric BREHAULT',
      author_email='eric.brehault@makina-corpus.org',
      url='http://plone.org/products/csvreplicata',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



# -*- coding: utf-8 -*-
#
# File: browser.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# GNU General Public License (GPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'

from Products.Five import BrowserView
from Products.CMFCore.utils import getToolByName
from zipfile import ZipFile
from cStringIO import StringIO

from Products.csvreplicata.interfaces import Icsvreplicata

import logging
logger = logging.getLogger('CSV MANAGER')

class ReplicationManager(BrowserView):
    """ allow to manage CSV replication
    """
    
    
    def __init__(self, context, request):
        """Initialize browserview"""
        self.context = context
        self.request = request
        
    def doImport(self):
        """
        """
        replicator = Icsvreplicata(self.context)
        datetimeformat = self.request.get("datetimeformat")
        vocabularyvalue = self.request.get("vocabularyvalue")
        encoding = self.request.get("encoding")
        delimiter = self.request.get("delimiter")
        stringdelimiter = self.request.get("stringdelimiter")
        conflict_winner = self.request.get("conflictrule")
        wf_transition = self.request.get("wf_transition")
        if wf_transition=="None":
            wf_transition = None
        importfiles = self.request.get("importfiles")
        if importfiles=="Yes":
            zip = ZipFile(self.request.get("csvfile"))
            csvfile=StringIO()
            csvfile.write(zip.read("export.csv"))
            csvfile.seek(0)
        else:
            zip = None
            csvfile = self.request.get("csvfile") 
        
        (count_created, count_modified, export_date, errors)=replicator.csvimport(csvfile, encoding=encoding, delimiter=delimiter, stringdelimiter=stringdelimiter, datetimeformat=datetimeformat, conflict_winner=conflict_winner, wf_transition=wf_transition, zip=zip, vocabularyvalue=vocabularyvalue)
        if len(errors)==0:
            self.writeMessageOnPage(["All lines imported, %d object(s) created, %d object(s) modified." % (count_created, count_modified)])
            self.writeMessageOnPage(["This CSV file has been produced on "+str(export_date)+". To avoid inconsistencies, do not import this file anymore. It is preferable to export a new one."])
        else:
            self.writeMessageOnPage(["%d object(s) created, %d object(s) modified." % (count_created, count_modified)])
            self.writeMessageOnPage(errors, error = True)
            
        self.request.RESPONSE.redirect(self.context.absolute_url()+'/@@csvreplicata')

    def doExport(self):
        """
        """
        replicator = Icsvreplicata(self.context)
        encoding = self.request.get('encoding')
        depth = int(self.request.get('depth','1'))
        datetimeformat = self.request.get("datetimeformat")
        vocabularyvalue = self.request.get("vocabularyvalue")
        delimiter = self.request.get("delimiter")
        stringdelimiter = self.request.get("stringdelimiter")
        wf_states = self.request.get("wf_states", 'Any')
        selected_exportable_content_types = self.request.get("exportable_content_types")

        # handling of selected types
        exportable_content_types = selected_exportable_content_types
        
        if 'Any' in wf_states:
            wf_states = None
        exportfiles = self.request.get("exportfiles")
        
        if exportfiles=="Yes":
            # initialize zipfile
            sIO = StringIO()
            zip = ZipFile(sIO,"w",8)
        else:
            zip = None
        
        csv = replicator.csvexport(encoding=encoding, delimiter=delimiter, stringdelimiter=stringdelimiter, depth=depth, datetimeformat=datetimeformat, wf_states=wf_states, zip=zip, vocabularyvalue=vocabularyvalue, exportable_content_types=exportable_content_types)
        
        if exportfiles=="Yes":
            self.request.response.setHeader('Content-type','application/zip')
            self.request.response.setHeader('Content-Disposition', "attachment; filename=export.zip")
        
            zip.writestr("export.csv",csv)
            zip.close()
            return sIO.getvalue()
        else:
            self.request.response.setHeader('Content-type','text/csv;charset='+encoding)
            self.request.response.setHeader('Content-Disposition', "attachment; filename=export.csv")
            return csv
        
    
    def writeMessageOnPage(self, messages, ifMsgEmpty = '', error = False):
        """adds portal message        
        """
        plone_tools = getToolByName(self, 'plone_utils')
        if plone_tools is not None:
            #portal message type
            if error:
                msgType = 'error'
            else:
                msgType = 'info'
            
            #if empty
            if (len(messages)==0):
               messages = [ifMsgEmpty]
            
            #display it
            for msg in messages:
                if len(msg)==0:
                    mess = ifMsgEmpty
                else:
                    mess = msg
                if len(mess)>0:
                    plone_tools.addPortalMessage(mess, msgType, self.request)
    
    def getAllWorkflowStates(self):
        """
        """
        plone_wf = getToolByName(self.context, 'portal_workflow')
        states = []
        for wf in plone_wf.getChildNodes():
            for s in wf.states.getChildNodes():
                if not(s.id in states):
                    states.append(s.id)
        return states
    
    def getAllWorkflowTransitions(self):
        """
        """
        plone_wf = getToolByName(self.context, 'portal_workflow')
        transitions = []
        for wf in plone_wf.getChildNodes():
            for t in wf.transitions.getChildNodes():
                if not(t.id in transitions):
                    transitions.append(t.id)
        return transitions


# -*- coding: utf-8 -*-
#
# File: setuphandlers.py
#
# Copyright (c) 2008 by ['Eric BREHAULT']
# Generator: ArchGenXML Version 2.0
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#

__author__ = """Eric BREHAULT <eric.brehault@makina-corpus.org>"""
__docformat__ = 'plaintext'


import logging
logger = logging.getLogger('csvreplicata: setuphandlers')
from Products.csvreplicata.config import PROJECTNAME
from Products.csvreplicata.config import DEPENDENCIES
import os
from Products.CMFCore.utils import getToolByName
import transaction
##code-section HEAD
##/code-section HEAD

def isNotcsvreplicataProfile(context):
    return context.readDataFile("csvreplicata_marker.txt") is None

def setupHideToolsFromNavigation(context):
    """hide tools"""
    if isNotcsvreplicataProfile(context): return 
    # uncatalog tools
    shortContext = context._profile_path.split(os.path.sep)[-3]
    if shortContext != 'csvreplicata': # avoid infinite recursions
        return
    site = context.getSite()
    toolnames = ['portal_csvreplicatatool']
    portalProperties = getToolByName(site, 'portal_properties')
    navtreeProperties = getattr(portalProperties, 'navtree_properties')
    if navtreeProperties.hasProperty('idsNotToList'):
        for toolname in toolnames:
            try:
                portal[toolname].unindexObject()
            except:
                pass
            current = list(navtreeProperties.getProperty('idsNotToList') or [])
            if toolname not in current:
                current.append(toolname)
                kwargs = {'idsNotToList': current}
                navtreeProperties.manage_changeProperties(**kwargs)

def fixTools(context):
    """do post-processing on auto-installed tool instances"""
    site = context.getSite()

    tool_ids=['portal_csvreplicatatool']
    for tool_id in tool_ids:
        tool=site[tool_id]
        tool.initializeArchetype()



def updateRoleMappings(context):
    """after workflow changed update the roles mapping. this is like pressing
    the button 'Update Security Setting' and portal_workflow"""
    if isNotcsvreplicataProfile(context): return 
    shortContext = context._profile_path.split(os.path.sep)[-3]
    if shortContext != 'csvreplicata': # avoid infinite recursions
        return
    wft = getToolByName(context.getSite(), 'portal_workflow')
    wft.updateRoleMappings()

def postInstall(context):
    """Called as at the end of the setup process. """
    # the right place for your custom code
    if isNotcsvreplicataProfile(context): return
    shortContext = context._profile_path.split(os.path.sep)[-3]
    if shortContext != 'csvreplicata': # avoid infinite recursions
        return
    site = context.getSite()



##code-section FOOT
##/code-section FOOT
